#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "audio.h"
#include "vmachine.h"
#include "config.h"
#include "vdc.h"
#include "allegro.h"
#include "cpu.h"
#include "debug.h"


long filesize(FILE *stream)
{
   long curpos, length;

   curpos = ftell(stream);
   fseek(stream, 0L, SEEK_END);
   length = ftell(stream);
   fseek(stream, curpos, SEEK_SET);
   return length;
}

void load_bios() {
	FILE *fn;

	fn=fopen("o2rom.bin","rb");
	if (!fn) {
		printf("Error loading bios rom o2rom.bin\n");
		exit(0);
	}
	fread(&rom1[0x0000],1,1024,fn);
	fclose(fn);
	memcpy(&rom2[0],&rom1[0],1024);
	memcpy(&rom3[0],&rom1[0],1024);
	memcpy(&rom4[0],&rom1[0],1024);
	printf("Bios ROM loaded ok\n");
}

void load_cart(char *file) {
	FILE *fn;
	long l;

	fn=fopen(file,"rb");
	if (!fn) {
		printf("Error loading %s\n",file);
		exit(0);
	}
	printf("Loading: %s Size: ",file);
	l=filesize(fn);
	if (l < 2049) {
		app_data.bank=1;
		fread(&rom1[1024],1,2048,fn);
		memcpy(&rom2[0],&rom1[0],3072);
		rom=rom1;
		printf("2K\n");
	}
	if (l == 4096) {
		app_data.bank=2;
		fread(&rom1[1024],1,2048,fn);
		fread(&rom2[1024],1,2048,fn);
		rom=rom2;
		printf("4K\n");
	}
	if (l == 8192) {
		app_data.bank=3;
		fread(&rom1[1024],1,2048,fn);
		fread(&rom2[1024],1,2048,fn);
		fread(&rom3[1024],1,2048,fn);
		fread(&rom4[1024],1,2048,fn);
		rom=rom4;
		printf("8K\n");
	}
	fclose(fn);
}


int main(int argc, char **argv)
{
	int i;
	char file[50];
	char attr[50],val[50];
	char *p;

	i=clock();
	
	//clrscr();
	printf("                         Virtural O2 v0.65\n");
	printf("                      copyright 1996/1998 by Daniel Boris\n");
	printf("\nVirtual O2 starting...\n");
	if (argc < 2) {
		printf("\nUse: O2 [file] [options]\n");
		printf("-debug	  Start the emulator in debug mode.\n");
		printf("-s1=n     Stick 1 mode.\n");
		printf("-s2=n     Stick 2 mode.\n");
		printf("    0 = Disable\n");
		printf("    1 = Right Keyboard Controls\n");
		printf("    2 = Left Keyboard Controls\n");
		printf("    3 = Joystick\n");
		printf("-nolimit  Turn off speed limiter.\n");
		printf("-nosound  Turn off sound emulation.\n");
		printf("-scanlines   Turn on Scanline mode.\n");
		exit(0);
	}
	app_data.debug = 0;
	app_data.stick1=1;
	app_data.stick2=2;
	app_data.bank=0;
    app_data.limit=1;
    app_data.sound_en=1;
	if (argc > 1) {
		for(i=1; i<argc; i++) {
			if (argv[i][0] != 45) 	{
				strcpy(file,argv[i]);
			}
			else
			{
				p=strtok(argv[i],"=");
				strcpy(attr,p);
				p=strtok(NULL,"=");
				if (p) strcpy(val,p);
				strlwr(attr);
				
				if (!strcmp(attr, "-scanlines")) {
					app_data.scan=1;
				}
				if (!strcmp(attr,"-nolimit")) {
					app_data.limit=0;
				}
				if (!strcmp(attr,"-nosound")) {
					app_data.sound_en=0;
				}

				if (!strcmp(attr,"-debug")) {
					app_data.debug=1;
				}

				if (!strcmp(attr,"-s1")) {
					sscanf(val,"%d",&app_data.stick1);
				}
				if (!strcmp(attr,"-s2")) {
					sscanf(val,"%d",&app_data.stick2);
				}
			}
		}
	}
	allegro_init();
    init_audio();
	printf("Using Allegro %s\n",allegro_id);
	load_bios();
	load_cart(file);

	//printf("\n\nPress any key to start.\n");
	//while(!kbhit()){};
	//getch();

    i=clock();			/* seed random number */
    srandom(i); 

	init_display();
	init_cpu();
	init_system();
	if (app_data.debug)
		debug();
	else
		run();
	exit(0);
}





