#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include <stdio.h>
#include "types.h"
#include "vmachine.h"
#include "config.h"
#include "allegro.h"
#include "keyboard.h"

extern	BYTE new_int;	/* Is new interrupt installed */

long colortable[24] =
{
  /* Background,Grid Dim */
  0x000000, 
  0x0000FF,   /* Blue */
  0x008000,   /* DK Green */
  0xff9b60,
  0xCC0000,   /* Red */
  0xa980ff, 
  0x82fddb, 
  0xFFFFFF,

  /* Background,Grid Bright */

  0x808080, 
  0x50AEFF,   /* Blue */
  0x00FF00,   /* Dk Green */
  0x82fbdb,   /* Lt Grey */
  0xEC0260,   /* Red */
  0xa980ff,   /* Violet */
  0xff9b60,   /* Orange */
  0xFFFFFF,

  /* Character,Sprite colors */
  0x717171,   /* Dark Grey */
  0xFF8080,   /* Red */
  0x00C000,   /* Green */
  0xff9b60,   /* Orange */
  0x50AEFF,   /* Blue */
  0xa980ff,   /* Violet */
  0x82fbdb,   /* Lt Grey */
  0xffffff    /* White */
};

#define VGA256		0x13
#define TEXT_MODE	0x03
#define PALETTE_MASK      0x3c6
#define PALETTE_REGISTER  0x3c8
#define PALETTE_DATA	  0x3c9

long clip_low;
long clip_high;

BITMAP *bmp;

void create_cmap(void);
void putvid(unsigned int ad, BYTE d, BYTE c);
void init_display();
void draw_grid();
void draw_display();
void draw_char(BYTE ypos,BYTE xpos,BYTE chr,BYTE col);
void set_textmode(void);

/* Collision buffer */
BYTE *col = NULL;

#define COL_SP0  0x01
#define COL_SP1  0x02
#define COL_SP2  0x04
#define COL_SP3  0x08
#define COL_VGRID 0x10
#define COL_HGRID 0x20
#define COL_CHAR 0x80

BYTE coltab[256];

struct AlColor
  {
    int red, green, blue;
  };

PALLETE colors;
PALLETE oldcol;

/* The pointer to the graphics buffer */
BYTE *vscreen = NULL;
extern BYTE cset[];

#define X_START     8
#define Y_START		24


void draw_region(void) {
	int i;

	i = master_clk/(LINECNT-1)-5;
	if (i<0) i=0;
//	printf("%d-%d, %x\n",last_line,i,VDCwrite[0xA3]); 
 
	clip_low = last_line * (long)320;
	clip_high = i * (long)320;
	if (clip_high > 80000) clip_high = 80000;
	if (clip_low < 0) clip_low=0; 
/*	printf("Low: %ld  High: %ld\n",clip_low,clip_high); */
	draw_display();
	last_line=i;
}

void create_cmap (void)
{
  int i;

  /* Initialise parts of the colors array */
  for (i = 0; i < 24; i++)
    {
      /* Use the color values from the color table */
      colors[i].r = (colortable[i] & 0xff0000) >> 18;
      colors[i].g = (colortable[i] & 0x00ff00) >> 10;
      colors[i].b = (colortable[i] & 0x0000ff) >> 2;
    }
}

void grmode() {
	if(app_data.scan) {
	set_gfx_mode(GFX_BEOS_WINDOWED,640,480,0,0);
	}
	else {
	set_gfx_mode(GFX_BEOS_WINDOWED,320,200,0,0);
	}
	set_display_switch_mode(SWITCH_BACKGROUND);
	set_palette_range(colors,0,256,0);
	init_keyboard();
}


void set_textmode(void) {
	set_palette_range(oldcol,0,256,0);
	set_gfx_mode(GFX_TEXT,80,25,0,0);
	if (new_int) Set_Old_Int9();
}

inline void putvid(unsigned int ad, BYTE d, BYTE c) {
	if (ad > clip_low && ad < clip_high) {
		vscreen[ad]=d;
		col[ad] |= c;
		coltab[c] |= col[ad];
		
	}
	
}

void draw_grid() {
	unsigned int pnt, pn1;
	BYTE mask,d;
	int j,i,x,k,w;
	BYTE color;
	BYTE bcolor;

	pnt=0;
	mask=0x01;
	color=VDCwrite[0xA3] & 0x07;
	color |= ((VDCwrite[0xA3] & 0x40) >> 3);
	for(j=0; j<9; j++) {
		pnt = (((j*24)+24) * 320);
		for (i=0; i<9; i++) {
			pn1 = pnt + (i * 32);
			d=VDCwrite[0xC0 + i];
			if (j == 8) {
				d=VDCwrite[0xD0];
				mask=1;
			}
			if (d & mask)	{
				for(x=0; x<36; x++) {
					putvid(pn1,color,COL_HGRID);
					putvid(pn1+320,color,COL_HGRID);
					putvid(pn1+640,color,COL_HGRID);
					pn1++;
				}
			}
		}
		mask = mask << 1;
	}
	mask=0x01;
	w=4;
	if (VDCwrite[0xA0] & 0x80) w=32;
	for(j=0; j<10; j++) {
		pnt=(j*32);
		mask=0x01;
		d=VDCwrite[0xE0+j];
		for (x=0; x<8; x++) {
			pn1 = pnt + (((x*24)+24) * 320);
			if (d & mask) {
				for(i=0; i<26; i++) {
					for (k=0; k<w; k++) {
						putvid(pn1+k,color,COL_VGRID);
					}
					pn1+=320;

				}
			}
			mask = mask << 1;
		}
	}
}

void finish_display() {
/*	if (app_data.debug) {
		setmode(VGA256);
		create_cmap();
	} */
	int h;
	if(app_data.scan) {
	for(h = 0; h < 250;h+=1)
	stretch_blit(bmp, screen, 0 , h, 310, 1, 0, (((480/2)-250))+(h << 1),640, 1);
    }
    else {
    blit(bmp,screen,0,24,10,0,310,200);
    }

/*	if (app_data.debug) {
		while(!kbhit()) {};
		getch();
		setmode(TEXT_MODE);
	} */

	memset(col,0,80000);
	coltab[0x01]=coltab[0x02]=0;
	coltab[0x04]=coltab[0x08]=0;
	coltab[0x10]=coltab[0x20]=0;
	coltab[0x40]=coltab[0x80]=0;

}

void draw_display() {
int i,j;
BYTE y,x,b,d1,cl,c;
unsigned int pnt,pnt2;
BYTE bcolor;


	bcolor=(VDCwrite[0xA3] & 0x38) >> 3;
	for (pnt=clip_low; pnt<clip_high; pnt++) vscreen[pnt]=bcolor;
	
	if (VDCwrite[0xA0] & 0x08) draw_grid();

	for(i=0x10; i<0x40; i+=4) {
		draw_char(VDCwrite[i],VDCwrite[i+1],VDCwrite[i+2],VDCwrite[i+3]);
	}
	pnt=0x40;
	for(i=0; i<4; i++) {
		x = VDCwrite[pnt+1];
		y = VDCwrite[pnt];
		for(j=0; j<4; j++) {
			draw_char(y,x,VDCwrite[pnt+2],VDCwrite[pnt+3]);
			x+=16;
			pnt+=4;
		}
	}
	pnt2=0x80;
	c=0x01;
	for (i=0; i<0x10; i+=4) {
		pnt2 = 0x80 + (i * 2);
		y=VDCwrite[i];
		x=VDCwrite[i+1]-8;
		cl=((VDCwrite[i+2] & 0x38) >> 3) + 16;
		if (x < 164 && y < 224 && y > 0	) {
			pnt = y * 320 + (x * 2);
			if (VDCwrite[i+2] & 0x04) {
				for (j=0; j<8; j++) {
					d1 = VDCwrite[pnt2++];
					for (b=0; b<8; b++) {
						if (d1 & 0x01) {
							putvid(pnt++,cl,c);
							putvid(pnt++,cl,c);
							putvid(pnt++,cl,c);
							putvid(pnt++,cl,c);
							putvid(pnt+316,cl,c);
							putvid(pnt+317,cl,c);
							putvid(pnt+318,cl,c);
							putvid(pnt+319,cl,c);
							putvid(pnt+636,cl,c);
							putvid(pnt+637,cl,c);
							putvid(pnt+638,cl,c);
							putvid(pnt+639,cl,c);
							putvid(pnt+956,cl,c);
							putvid(pnt+957,cl,c);
							putvid(pnt+958,cl,c);
							putvid(pnt+959,cl,c);

						}
						else {
							pnt+=4;
						}
						d1 = d1 >> 1;
					}
					pnt=pnt+1248;
				}

			}
			else {
				for (j=0; j<8; j++) {
					d1 = VDCwrite[pnt2++];
					for (b=0; b<8; b++) {
						if (d1 & 0x01) {
							putvid(pnt++,cl,c);
							putvid(pnt++,cl,c);
							putvid(pnt+318,cl,c);
							putvid(pnt+319,cl,c);
						}
						else {
							pnt+=2;
						}
						d1 = d1 >> 1;
					}
					pnt=pnt+624;
				}
			}
		}
		c = c << 1;
	}

}

void draw_char(BYTE ypos,BYTE xpos,BYTE chr,BYTE col) {
	int j,c;
	BYTE cl,d1;
	int y,b,n;
	unsigned int pnt;

	y=(ypos & 0xFE); 
	ypos = ypos >> 1;
	if (y < 214) {
		c=(int)chr + ypos;
		if (col & 0x01) c+=256;
		if (c > 511) c=c-512;

		n = 8 - (ypos % 8) - (chr % 8);
		if (n < 3) n = n + 7;
		cl=((col & 0x0E) >> 1) + 16;
		pnt = y * 320 + ((xpos-8) * 2);
		for (j=0; j<n; j++) {
			d1 = cset[c+j];
			for (b=0; b<8; b++) {
				if (d1 & 0x80) {
					putvid(pnt++,cl,COL_CHAR);
					putvid(pnt++,cl,COL_CHAR);
					putvid(pnt+318,cl,COL_CHAR);
					putvid(pnt+319,cl,COL_CHAR);
				}
				else {
					pnt+=2;
				}
				d1 = d1 << 1;
			}
			pnt=pnt+624;
		}
	}
}

void close_display() {
	free(vscreen);
	free(col);
}

void init_display() {
	
	get_palette_range(oldcol,0,256);
	create_cmap ();

	bmp=create_bitmap(320,250);
	vscreen=(BYTE *) bmp->dat;
	if(!vscreen) {
		printf("Could not allocate memory for screen buffer.\n");
		exit(0);
	}
    clear(bmp);
	col=(BYTE *)malloc(80000);
	if (!col) {
		printf("Could not allocate memory for collision buffer.\n");
		free(vscreen);
		exit(0);
	}
	memset(col,0,80000);

	if (!app_data.debug) {
		grmode();
	}

}
