#include <stdlib.h>
#include <stdio.h>
#include <dos.h>
#include <time.h>

#include "audio.h"
#include "types.h"
#include "cpu.h"
#include "keyboard.h"
#include "config.h"
#include "debug.h" 
#include "allegro.h"
#include "vdc.h" 

struct resource {
	int scan;
	int bank;
	int debug;
	int stick1;
	int stick2;
	int limit;
	int sound_en;
	char path[80];
} app_data;

BYTE stick1,stick2;

int frame=0;
int clock_start,clk_end;


#define LINECNT 	21

unsigned int key_map[6][8]= {{11,2,3,4,5,6,7,8},
	{9,10,0,0,57,53,38,25},
	{78,17,18,19,20,22,23,24},
	{16,31,32,33,34,35,36,37},
	{30,44,45,46,47,48,50,52},
	{12,55,39,13,21,49,14,28}};

int last_line;
BYTE x_latch,y_latch;
BYTE line_count;
int h_clk;			/* horizontal clock */
int master_clk;		/* Master clock */
int master_count;
BYTE rom1[5120];
BYTE rom2[5120];
BYTE rom3[5120];
BYTE rom4[5120];

BYTE intRAM[64];
BYTE extRAM[256];
BYTE VDCwrite[256];
BYTE *rom;

BYTE state;
#define ST_DISPLAY 0
#define ST_VBLANK 1

BYTE read_P2(void);
void ext_write(BYTE dat, ADDRESS adr);
BYTE ext_read(ADDRESS adr);
void handle_clock();
void draw_char(BYTE ypos,BYTE xpos,BYTE chr,BYTE col);


void run() {
	int i,j;

	while(!key_done) {
		exec();
	}
	set_textmode();
	app_data.debug=1;
	if (key_debug) {
		close_audio();
		debug();
	}
	close_display();
	close_audio();
	exit(0);
}

void handle_clock() {
  	long i;
	clock_t curr;
	static clock_t last;

	master_clk+=clk;
	h_clk+=clk;

	if (h_clk > LINECNT-1) {
		h_clk-=LINECNT;
//        if (VDCwrite[0xA0] & 0x01) ext_IRQ();
		if (count_on && state == 1) {
			timer++;
			if (timer == 0) {
				tim_IRQ();
				t_flag=1;
				draw_region();
			}
		}
	}

	if (timer_on) {
		master_count+=clk;
		if (master_count > 31) {
			master_count-=31;
			timer++;
			if (timer == 0) {
				tim_IRQ();
			}
		}
	}

	if (master_clk > 5493) {
		if (state == 1) {
			handle_key();
			update_audio();
//			printf("NEW FRAME\n");
			draw_region();
			ext_IRQ();
			state = 2;
		}
		else
		{
			if (master_clk > 5600 && state == 2) {
				state = 3;
			}
			if (master_clk > 5970) {
				state=1;
				last_line=0;
				master_clk-=5970;
			    frame++;
				if (app_data.limit) {
					do {
						curr = clock();
						if (curr - last == 0) curr++;				
					} while (CLOCKS_PER_SEC / (curr - last) > 60 * 11 /10);
					last = curr;
				}
				if (!app_data.debug) {
					finish_display();
				}
			}
		}
	}
}

void init_system() {
	int i,j;

	last_line=0;
	stick1=0x00;
	stick2=0x00;
	state=ST_DISPLAY;
	master_clk=0;
	h_clk=0;
	line_count=0;
	timer=0;
	for(i=0; i<256; i++) {
		VDCwrite[i]=0;
		extRAM[i]=0;
	}
	for(i=0; i<64; i++) {
		intRAM[i]=0;
	}
	if (app_data.stick1 == 3 || app_data.stick2 == 3) {
		install_joystick(JOY_TYPE_AUTODETECT);
		i = initialise_joystick();
		if (i) {
			printf("Hello\n");
			if (app_data.stick1 == 2) app_data.stick1 = 1;
			if (app_data.stick2 == 2) app_data.stick2 = 1;
		}
	}
}

BYTE read_t1(void) {
	if (h_clk > 16)
		return 1;
	else
		return 0;
}


void write_p1(BYTE d) {
	p1 = d;
	if (app_data.bank == 2) {
		if (p1 & 0x01)
			rom=rom2;
		else
			rom=rom1;
	}
	if (app_data.bank == 3) {
		switch(p1 & 0x03) {
			case 0:
				rom=rom1;
				break;
			case 1:
				rom=rom2;
				break;
			case 2:
				rom=rom3;
				break;
			case 3:
				rom=rom4;
				break;
		}
	}
}

BYTE read_P2(void) {
int i,si,so;

	if (!(p1 & 0x04)) {
		if (p2 & 0x04)
			si = (p2 & 0x03) + 4;
		else
			si = (p2 & 0x03);

		so=0xff;
		for (i=0; i<8; i++) {
			if (key[ key_map[si][i] ]) {
				so = i ^ 0x07;
			}
		}

		if (so != 0xff) {
			p2 = p2 & 0x0F;
			p2 = p2 | (so << 5);
		}
		else
		{
			p2 = p2 | 0xF0;
		}
	 }
	 else {
		p2 = p2 | 0xF0;
	 }
	 return(p2);
}

BYTE ext_read(ADDRESS adr) {
	BYTE d;
	BYTE si;
	BYTE m,m2;
	int i,j;


	if (!(p1 & 0x08) && !(p1 & 0x40)) {		/* Handle VDC Read */
		switch(adr) {
		case 0xA1:
			d = VDCwrite[0xA0] & 0x02;
//			if (state == 2) d = d | 0x08;
			if (master_clk > 5493) d = d | 0x08;
            if (h_clk < (LINECNT-7)) d = d | 0x01;
            if (sound_IRQ) d = d | 0x04;
            sound_IRQ=0;
			return d;
		case 0xA2:
			si = VDCwrite[0xA2];
			m=0x01;
			d=0;
			for(i=0; i<8; i++) {
				if (si & m) {
					if (coltab[1] & m) d = d | (coltab[1] & (m ^ 0xFF));
					if (coltab[2] & m) d = d | (coltab[2] & (m ^ 0xFF));
					if (coltab[4] & m) d = d | (coltab[4] & (m ^ 0xFF));
					if (coltab[8] & m) d = d | (coltab[8] & (m ^ 0xFF));
					if (coltab[0x10] & m) d = d | (coltab[0x10] & (m ^ 0xFF));
					if (coltab[0x20] & m) d = d | (coltab[0x20] & (m ^ 0xFF));
					if (coltab[0x80] & m) d = d | (coltab[0x80] & (m ^ 0xFF));
				}
				m = m << 1;
			}

			return d;
		case 0xA5:
			if (!(VDCwrite[0xA0] & 0x02)) {
				return x_latch;
			}
			else {
				x_latch = h_clk * 12;
				return x_latch;
			}			
		case 0xA4:
			if (!(VDCwrite[0xA0] & 0x02)) {
				return y_latch;
			}
			else {
				y_latch = master_clk/22;
				if (y_latch > 241) y_latch=0xFF;
				return y_latch;
			}
		default:
			return VDCwrite[adr];
		}
	}

	if (!(p1 & 0x10)) {		/* Handle ext RAM Read */
		 return extRAM[adr & 0xFF];
	}

	return 0;
}

BYTE in_bus(void) {
	BYTE si,d,mode;

	if ((p1 & 0x08) && (p1 & 0x10)) {       /* Handle joystick read */
		if (!(p1 & 0x04)) {
			if (p2 & 0x04)
				si = (p2 & 0x03) + 4;
			else
				si = (p2 & 0x03);
		}
		d=0xFF;
		if (si == 1) 
			mode = app_data.stick1;
		else
			mode = app_data.stick2;
		switch(mode) {
			case 1:
				d=keyjoy1();
				break;
			case 2:
				d=keyjoy2();
				break;
			case 3: 
				poll_joystick();
				if (joy_up) d &= 0xFE;
				if (joy_right) d &= 0xFD;
				if (joy_down) d &= 0xFB;
				if (joy_left) d &= 0xF7;
				if (joy_b1) d &= 0xEF;
				break;
		}
		if (stick1) d = stick1;
   }
   return d;
}

void ext_write(BYTE dat, ADDRESS adr) {
	int i;

	if (!(p1 & 0x08)) {		/* Handle VDC Write */		
		if (adr == 0xA0) {
			if ((VDCwrite[0xA0] & 0x02) && !(dat & 0x02)) {
				y_latch = master_clk/22;
				x_latch = h_clk * 12;
				if (y_latch > 241) y_latch=0xFF;
			}
		}
		if (adr > 0xA8 && adr < 0xAB) {
//			printf("SOUND: %x = %x\n",adr,dat);
		}
		VDCwrite[adr]=dat;
	}

	if (!(p1 & 0x10) && !(p1 & 0x40)) {		/* Handle ext RAM Write */
		extRAM[adr & 0xFF] = dat;
	}

}

